// Copyright 2004 Adam Megacz, see the COPYING file for licensing [GPL]
package org.ibex.core;

import java.io.*;
import org.ibex.js.*;
import org.ibex.util.*;
import org.ibex.graphics.*;
import org.ibex.plat.*;
import org.ibex.net.*;
import org.ibex.crypto.*;

/** Singleton class that provides all functionality in the ibex.* namespace */
public final class Ibex extends JS implements JS.Cloneable {

    // FIXME remove this
    private final JS rr;

    public Ibex(Stream rr) { try { this.rr = bless(rr);} catch(JSExn e) { throw new Error("should never happen: " + e); } }

    public JS resolveString(String str, boolean permitAbsolute) throws JSExn {
        if (str.indexOf("://") != -1) {
            if (permitAbsolute) return url2res(str);
            throw new JSExn("absolute URL " + str + " not permitted here");
        }
        // root-relative
        //JS ret = (JS)getAndTriggerTraps("");
        //FIXME
        JS ret = rr;
        while(str.indexOf('.') != -1) {
            String path = str.substring(0, str.indexOf('.'));
            str = str.substring(str.indexOf('.') + 1);
            ret = ret.get(JS.S(path));
        }
        if (!"".equals(str)) ret = ret.get(JS.S(str));
        return ret;
    }

    /** lets us put multi-level get/put/call keys all in the same method */
    private class Sub extends JS {
        String key;
        Sub(String key) { this.key = key; }
        public void put(JS key, JS val) throws JSExn { Ibex.this.put(this.key + "." + JS.toString(key), val); }
        public JS get(JS key) throws JSExn { return Ibex.this.get(this.key + "." + JS.toString(key)); }
        public JS call(JS a0, JS a1, JS a2, JS[] rest, int nargs) throws JSExn {
            return Ibex.this.callMethod(this.key, a0, a1, a2, rest, nargs);
        }
        public JS callMethod(JS method, JS a0, JS a1, JS a2, JS[] rest, int nargs) throws JSExn {
            return Ibex.this.callMethod(this.key + "." + JS.toString(method), a0, a1, a2, rest, nargs);
        }
    }
    private Cache subCache = new Cache(20);
    private Sub getSub(String s) {
        Sub ret = (Sub)subCache.get(s);
        if (ret == null) subCache.put(s, ret = new Sub(s));
        return ret;
    }

    public JS get(JS key) throws JSExn { return JS.isString(key) ? get(JS.toString(key)) : null; }
    JS get(String name) throws JSExn {
        if (name instanceof String && ((String)name).length() == 0) return rr;
        //#switch(name)
        case "math": return ibexMath;
        case "string": return ibexString;
        case "date": return METHOD;
        case "box": return new Box();
        case "clone": return METHOD;
        case "bless": return METHOD;
        case "regexp": return METHOD;
        case "ui": return getSub("ui");
        case "ui.font": return getSub("ui.font");
        case "ui.font.wait": return METHOD;
        case "ui.font.width": return METHOD;
        case "ui.font.height": return METHOD;
        case "ui.font.sansserif": return Main.builtin.get(JS.S("fonts/vera/Vera.ttf"));
        case "ui.font.monospace": return Main.builtin.get(JS.S("fonts/vera/VeraMono.ttf"));
        case "ui.font.serif": return Main.builtin.get(JS.S("fonts/vera/VeraSe.ttf"));
        case "ui.browser": return METHOD;
        case "ui.mouse": return getSub("ui.mouse");
        case "ui.mouse.button":
            if (Surface.button1 && !Surface.button2 && !Surface.button3) return N(1);
            else if (!Surface.button1 && Surface.button2 && !Surface.button3) return N(2);
            else if (!Surface.button1 && !Surface.button2 && Surface.button3) return N(3);
            else return ZERO;
        case "ui.key": return getSub("ui.key");
        case "ui.key.name": return getSub("ui.key.name");
        case "ui.key.name.alt": return JS.S(Platform.altKeyName());
        case "ui.key.alt": return Surface.alt ? T : F;
        case "ui.key.control": return Surface.control ? T : F;
        case "ui.key.shift": return Surface.shift ? T : F;
        case "ui.clipboard": return JS.S((String)Platform.getClipBoard());
        case "ui.maxdim": return N(Short.MAX_VALUE);
        case "ui.screen": return getSub("ui.screen");
        case "ui.screen.width": return N(Platform.getScreenWidth());
        case "ui.screen.height": return N(Platform.getScreenHeight());
        case "undocumented": return getSub("undocumented");
        case "undocumented.initialOrigin": return JS.S(Main.origin);
        case "undocumented.initialTemplate": return JS.S(Main.initialTemplate);
        case "thread": return getSub("thread");
        case "thread.yield": return METHOD;
        case "thread.sleep": return METHOD;
        case "stream": return getSub("stream");
        case "stream.homedir": return url2res("file:" + System.getProperty("user.home"));
        case "stream.tempdir": return url2res("file:" + System.getProperty("java.io.tempdir"));
        case "stream.watch": return METHOD;
        case "stream.unzip": return METHOD;
        case "stream.uncab": return METHOD;
        case "stream.cache": return METHOD;
        case "stream.url": return METHOD;
        case "stream.parse.html": return METHOD;
        case "stream.parse.xml": return METHOD;
        case "stream.parse.utf8": return METHOD;
        case "net": return getSub("net");
        case "net.rpc": return getSub("net.rpc");
        case "net.rpc.xml": return METHOD;
        case "net.rpc.soap": return METHOD;
        case "log": return getSub("log");
        case "log.debug": return METHOD;
        case "log.info": return METHOD;
        case "log.warn": return METHOD;
        case "log.error": return METHOD;
        case "crypto": return getSub("crypto");
        case "crypto.rsa": return METHOD;
        case "crypto.md5": return METHOD;
        case "crypto.sha1": return METHOD;
        case "crypto.rc4": return METHOD;
        //#end
        return null;
    }

    public void put(JS key, JS value) throws JSExn { if(JS.isString(key)) put(JS.toString(key),value); else super.put(key,value); }
    void put(Object name, final JS value) throws JSExn {
        //#switch(name)
        case "thread": Scheduler.add((Task)value); return;
        case "ui.clipboard": Platform.setClipBoard(JS.toString(value)); return;
        case "ui.frame": Platform.createSurface((Box)value, true, true); return;
        case "ui.window": Platform.createSurface((Box)value, false, true); return;
        case "undocumented.proxyAuthorization":
            HTTP.Proxy.Authorization.authorization = value.toString();
            HTTP.Proxy.Authorization.waitingForUser.release();
            return;
        //#end
        throw new JSExn("attempted to put unknown property: ibex."+name);
    }

    public JS callMethod(JS name, JS a, JS b, JS c, JS[] rest, int nargs) throws JSExn {
        if(JS.isString(name)) return callMethod(JS.toString(name),a,b,c,rest,nargs);
        return super.callMethod(name,a,b,c,rest,nargs);
    }
    JS callMethod(String name, JS a, JS b, JS c, JS[] rest, int nargs) throws JSExn {
        try {
            //#switch(name)
            case "date": return new JSDate(a, b, c, rest, nargs);
            case "net.rpc.soap": return new SOAP(JS.toString(a), "", JS.toString(b), JS.toString(c));
                // FIXME support object dumping
            case "log.debug":    JS.debug(a== null ? "**null**" : JS.debugToString(a)); return null;
            case "log.info":     JS.info(a== null ? "**null**" : JS.debugToString(a)); return null;
            case "log.warn":     JS.warn(a== null ? "**null**" : JS.debugToString(a)); return null;
            case "log.error":    JS.error(a== null ? "**null**" : JS.debugToString(a)); return null;
            //#end
 
            switch (nargs) {
                case 0:
                    //#switch(name)
                    case "thread.yield": sleep(0); return null;
                    //#end
                    break;
                case 1:
                    //#switch(name)
                    case "clone":
                        if(a == null) throw new JSExn("can't clone the null value");
                        return ((JS)a).jsclone();
                    case "bless": return bless((JS)a);
                    case "ui.browser": Platform.newBrowserWindow(JS.toString(a)); return null;
                    case "stream.unzip": return a == null ? null : new Stream.Zip(a);
                    case "stream.uncab": return a == null ? null : new Stream.Cab(a);
                    case "stream.cache":
                        try { return a == null ? null : new Stream.CachedStream(a, "resources", true); }
                        catch (Stream.NotCacheableException e) { throw new JSExn("this resource cannot be cached"); }
                    case "stream.url": {
                        String url = JS.toString(a);
                        if (url.startsWith("http://")) return new Stream.HTTP(url);
                        else if (url.startsWith("https://")) return new Stream.HTTP(url);
                        else if (url.startsWith("data:")) return new Stream.ByteArray(Base64.decode(url.substring(5)), null);
                        else if (url.startsWith("utf8:")) return new Stream.ByteArray(url.substring(5).getBytes(), null);
                        else if (url.startsWith("file:")) {
                            // FIXME
                            Platform.fileDialog(url.substring(5), false);
                        }
                        throw new JSExn("invalid resource specifier " + url);
                    }
                    case "thread.sleep": sleep(JS.toInt(a)); return null;
                    case "regexp": return new JSRegexp(a, null);
                    case "net.rpc.xml": return new XMLRPC(JS.toString(a), "");
                    case "crypto.rsa": /* FEATURE */ return null;
                    case "crypto.md5": /* FEATURE */ return null;
                    case "crypto.sha1": /* FEATURE */ return null;
                    case "crypto.rc4": /* FEATURE */ return null;
                    case "stream.parse.html": throw new JSExn("not implemented yet"); //return null;
                    case "stream.parse.xml": if(a == null) return null; new XMLHelper(b).doParse(a); return null;
                        // FIXME backgrounding
                    case "stream.parse.utf8": if(a == null) return null;
                                              try { return JS.S(new String(InputStreamToByteArray.convert(a.getInputStream()))); }
                                              catch (Exception e) { Log.warn(this, e); }
                    //#end
                    break;
                case 2:
                    //#switch(name)
                    case "stream.watch": return new Stream.ProgressWatcher(a, b);
                    case "regexp": return new JSRegexp(a, b);
                    //#end
                case 3:
                    //#switch(name)
                    case "ui.font.height": return N(Font.getFont(a, JS.toInt(b)).textheight(JS.toString(c)));
                    case "ui.font.wait": throw new Error("FIXME: ibex.ui.font.wait not implemented");
                    case "ui.font.width": return N(Font.getFont(a, JS.toInt(b)).textwidth(JS.toString(c)));
                    //#end
                    break;
            }
        } catch (RuntimeException e) {
            // FIXME: maybe JSExn should take a second argument, Exception
            Log.warn(this, "ibex."+name+"() threw: " + e);
            throw new JSExn("invalid argument for ibex object method "+name+"()");
        }

        throw new JSExn("invalid number of arguments ("+nargs+") for ibex object method "+name+"()");
    }

    public JS url2res(String url) throws JSExn {
        if (url.startsWith("http://")) return new Stream.HTTP(url);
        else if (url.startsWith("https://")) return new Stream.HTTP(url);
        else if (url.startsWith("data:")) return new Stream.ByteArray(Base64.decode(url.substring(5)), null);
        else if (url.startsWith("utf8:")) return new Stream.ByteArray(url.substring(5).getBytes(), null);
        else throw new JSExn("invalid resource specifier " + url);
        // FIXME support file:// via dialog boxes
    }

    public static void sleep(final int i) throws JSExn {
        try {
            final JS.UnpauseCallback callback = JS.pause();
            // FEATURE use a single sleeper thread
            new Thread() { public void run() {
                try { Thread.sleep(i); } catch (InterruptedException e) { }
                Scheduler.add(callback);
            } }.start();
        } catch (JS.NotPauseableException npe) {
            throw new JSExn("you cannot sleep or yield in the foreground thread");
        }
    }
    
    public static final JSMath ibexMath = new JSMath() {
            // FEATURE: find a cleaner way to do this
            private JS gs = /*new JSScope.Global();*/ null; // FIXME: Global scope
            public JS get(JS key) throws JSExn {
                //#switch(JS.toString(key))
                case "isNaN": return METHOD;
                case "isFinite": return METHOD;
                case "NaN": return METHOD;
                case "Infinity": return METHOD;
                //#end
                return super.get(key);
            }
            public JS callMethod(JS name, JS a, JS b, JS c, JS[] rest, int nargs) throws JSExn {
                //#switch(JS.toString(name))
                case "isNaN": return gs.callMethod(name,a,b,c,rest,nargs);
                case "isFinite": return gs.callMethod(name,a,b,c,rest,nargs);
                case "NaN": return gs.callMethod(name,a,b,c,rest,nargs);
                case "Infinity": return gs.callMethod(name,a,b,c,rest,nargs);
                //#end
                return super.callMethod(name,a,b,c,rest,nargs);
            }
    };

    public static final JS ibexString = new JS() {
            private JS gs = /*new JSScope.Global();*/ null; // FIXME: Global scope
            public JS get(JS key) throws JSExn {
                //#switch(JS.toString(key))
                case "parseInt": return METHOD;
                case "parseFloat": return METHOD;
                case "decodeURI": return METHOD;
                case "decodeURIComponent": return METHOD;
                case "encodeURI": return METHOD;
                case "encodeURIComponent": return METHOD;
                case "escape": return METHOD;
                case "unescape": return METHOD;
                case "fromCharCode": return METHOD;
                //#end
                return super.get(key);
            }
            public JS callMethod(JS name, JS a, JS b, JS c, JS[] rest, int nargs) throws JSExn {
                //#switch(JS.toString(name))
                case "parseInt": return gs.callMethod(name,a,b,c,rest,nargs);
                case "parseFloat": return gs.callMethod(name,a,b,c,rest,nargs);
                case "decodeURI": return gs.callMethod(name,a,b,c,rest,nargs);
                case "decodeURIComponent": return gs.callMethod(name,a,b,c,rest,nargs);
                case "encodeURI": return gs.callMethod(name,a,b,c,rest,nargs);
                case "encodeURIComponent": return gs.callMethod(name,a,b,c,rest,nargs);
                case "escape": return gs.callMethod(name,a,b,c,rest,nargs);
                case "unescape": return gs.callMethod(name,a,b,c,rest,nargs);
                case "fromCharCode": return gs.callMethod(name,a,b,c,rest,nargs);
                //#end
                return super.callMethod(name,a,b,c,rest,nargs);
            }
    };

    private class XMLHelper extends XML {
        private class Wrapper extends XML.Exn { public JSExn wrapee; public Wrapper(JSExn jse) { super(""); wrapee = jse; } }
        private JS characters, whitespace, endElement, startElement;
        public XMLHelper(JS b) throws JSExn {
            super(BUFFER_SIZE);
            startElement = b.getAndTriggerTraps(JS.S("startElement"));
            endElement   = b.getAndTriggerTraps(JS.S("endElement"));
            characters   = b.getAndTriggerTraps(JS.S("characters"));
            whitespace   = b.getAndTriggerTraps(JS.S("whitespace"));
        }

        public void startElement(XML.Element c) throws XML.Exn { try {
                JS attrs = new JS.O();
                // FIXME attribute URIs? add an additional hash?
                for(int i=0; i<c.getAttrLen(); i++) attrs.put(JS.S(c.getAttrKey(i)), JS.S(c.getAttrVal(i)));
                startElement.call(JS.S(c.getLocalName()), attrs, JS.S(c.getUri()), null, 3);
        } catch (JSExn jse) { throw new Wrapper(jse); } }

        public void endElement(XML.Element c) throws XML.Exn { try {
                endElement.call(JS.S(c.getLocalName()), JS.S(c.getUri()), null, null, 2);
        } catch (JSExn jse) { throw new Wrapper(jse); } }

        public void characters(char[] ch, int start, int length) throws XML.Exn { try {
                characters.call(JS.S(new String(ch, start, length)), null, null, null, 1);
        } catch (JSExn jse) { throw new Wrapper(jse); } }

        public void whitespace(char[] ch, int start, int length) throws XML.Exn { try {
                whitespace.call(JS.S(new String(ch, start, length)), null, null, null, 1);
        } catch (JSExn jse) { throw new Wrapper(jse); } }

        public void doParse(JS s) throws JSExn {
            try { 
                parse(new BufferedReader(new InputStreamReader(s.getInputStream())));
            } catch (Wrapper e) {
                throw e.wrapee;
            } catch (XML.Exn e) {
                throw new JSExn("error parsing XML: " + e.toString());
            } catch (IOException e) {
                if (Log.on) Log.info(this, "IO Exception while reading from file");
                if (Log.on) Log.info(this, e);
                throw new JSExn("error reading from Resource");
            }
        }
    }

    // FEATURE: move this into builtin.xwar
    public Blessing bless(JS b) throws JSExn { return new Ibex.Blessing(b, this, null, null); }
    // FIXME: Does this really need to extends JS.Clone?
    // FEATURE: Mandate that Blessings use only String keys?
    public static class Blessing extends JS.Clone {
        private Ibex ibex;
        private Template t = null;
        public JS parentkey = null;
        public Blessing parent = null;
        private Hash cache = new Hash();
        public Blessing(JS clonee, Ibex ibex, Blessing parent, JS parentkey) throws JSExn {
            super(clonee); this.ibex = ibex; this.parentkey = parentkey; this.parent = parent; }
        public JS get(JS key) throws JSExn {
            if (JS.isString(key) && JS.toString(key).equals("")) return getStatic();
            if (cache.get(key) != null) return (JS)cache.get(key);
            JS ret = new Blessing(clonee.get(key), ibex, this, key);
            cache.put(key, ret);
            return ret;
        }
        public static Blessing getBlessing(JS js) {
            while (js instanceof JS.Clone && !(js instanceof Blessing)) js = ((JS.Clone)js).getClonee();
            if (!(js instanceof Blessing)) return null;
            return (Blessing)js;
        }
        // FEATURE: This is a gross hack
        public InputStream getImage() throws JSExn {
            try {
                InputStream in = getInputStream();
                if (in != null) return in;
            } catch (IOException e) { /* DELIBERATE */ }
            String[] exts = new String[] { ".png", ".jpeg", ".gif" };
            for (int i=0; i < exts.length; i++)
                try {
                    InputStream in = parent.get(JS.S(JS.toString(parentkey) + exts[i])).getInputStream();
                    if (in != null) return in;
                } catch (IOException f) { /* DELIBERATE */ }
            return null;
        }
        public JSScope getStatic() throws JSExn {
            try {
                if (t == null) {
                    // FEATURE: Might want to handle the ".t" part better
                    JS res = parent.get(JS.S(JS.toString(parentkey) + ".t"));
                    t = Template.buildTemplate(description(), res, ibex);
                }
                return t != null ? t.staticScope : null;
            } catch (Exception e) {
                Log.error(this, e);
                return null;
            }
        }
        private String description() {
            String s = JS.debugToString(parentkey);
            for(Blessing b = parent; b != null; b = b.parent) s = JS.debugToString(parentkey) + "." + s;
            return s;
        }
        public JS call(JS a, JS b, JS c, JS[] rest, int nargs) throws JSExn {
            if (nargs != 1) throw new JSExn("can only call a template with one arg");
            getStatic();
            if (t == null) throw new JSExn("No such template " + parentkey);
            if(!(a instanceof Box)) throw new JSExn("can only apply templates to boxes");
            t.apply((Box)a);
            return a;
        }
        // FEATURE: This is a gross hack
        Template getTemplate()  throws JSExn {
            getStatic();
            if (t == null) throw new JSExn("No such template " + JS.debugToString(parentkey));
            return t;
        }
    }

}
